/* begincopyright
  Copyright (c) 1988 Xerox Corporation. All rights reserved.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted. Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws. This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification. Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    PCR Coordinator
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA
  endcopyright */
/*
 * ThreadsUIOStream.c
 *
 * Demers, March 30, 1990 0:30:39 am PST
 * Boehm, September 26, 1990 3:51:19 pm PDT
 *
 * Unix I/O for Xerox Runtime threads package, System V streams descriptors.
 */

 
#include "xr/ThreadsBackdoor.h"
#include "xr/ThreadsSignalsPrivate.h"
#include "xr/ThreadsMsgPrivate.h"
#include "xr/ThreadsStatsPrivate.h"
#include "xr/UIO.h"
#include "xr/UIOPrivate.h"
#include "xr/Errno.h"

#include <sys/types.h>
#include <sys/file.h>
#include <sys/fcntl.h>

#define XR_PROTECT_POLL_FROM_SIGTSTP 1	/* workaround SunOS 4.0.3 bug */

#if XR_PROTECT_POLL_FROM_SIGTSTP
#   include <sys/signal.h>
#endif


#undef XR_ASSERT
#define XR_ASSERT(b,s) if(!(b)) XR_Panic(s)


/*
 *
 * Descriptor utilities
 *
 */



int
XR_UIOIOPSetNonBlockingStream (d)
    int d;
{
#   if STREAMS
    int result, cFlags;

    if( (cFlags = fcntl(d, F_GETFL, NIL)) < 0 ) {
	result = -errno;
	goto Out;
    }
    cFlags |= FNBIO;
    if( fcntl(d, F_SETFL, cFlags) < 0 ) {
	result = -errno;
	goto Out;
    }
    result = d;

Out:
    return(result);
#   else
	XR_Panic("UIOIOPSetNonBlockingStream 0");
#   endif
}



int
XR_UIOPollStream(fds, nfds, timeout, flagp)
    XR_WaitReadyData fds;
    int nfds;
    unsigned timeout;
    unsigned *flagp;
{
#   if STREAMS
	int ans;

	if( nfds == 0 ) {
	    /* ??? this may be unnecessary ??? */
	    XR_SpinStep(timeout*1000); /* *1000 for msec to usec -mdw 8/6/89 */
            errno = EINTR;
            ans = (-1);
	} else if( (*flagp) ) {
	    errno = EINTR;
	    ans = (-1);
	} else {
#           if XR_PROTECT_POLL_FROM_SIGTSTP
                int mask = sigblock(sigmask(SIGTSTP));
#           endif
            ans = poll(fds, nfds, timeout);
            /* We assert XR_ProtectSysCall is inappropriate, since we are in */
            /* an IOP.							     */
            if( ans < 0 ) switch(errno) {
                case EINTR:
                case EAGAIN:
                    break;
                default:
                    XR_ConsoleMsg("%? fds 0x%x nfds %d \n", fds, nfds);
                    XR_Panic("UIOPollStream 0");
            }
#           if XR_PROTECT_POLL_FROM_SIGTSTP
                { int sav = errno; sigsetmask(mask); errno = sav; }
#           endif
        }
	return ans;
#   else
	XR_Panic("UIOPollStream 1");
	return(-1);
#   endif
}


/*
 * Exported to UIO.h
 */

int
XR_OpenStream (name, flags, mode)
    char *name;
    unsigned flags;
    unsigned mode;
{
    return( XR_Open4(name, flags, mode, XR_FD_KIND_STREAM) );
}



#if STREAMS

struct XR_GetMsgArgsRep {
    struct strbuf *gma_ctlptr;
    struct strbuf *gma_dataptr;
    int *gma_flags;
};

#define gma	((struct XR_GetMsgArgsRep *)(x1))

static XR_FDE_FDC_WORKER(XR_GetMsgWorker)
{
    struct strbuf myCtl;
    struct strbuf myData;
    struct strbuf *myCtlPtr, *myDataPtr;
    int myFlags;
    int * myFlagsPtr;
    int ans;

    if ( gma->gma_dataptr ) {
        myData = *(gma->gma_dataptr);
        myDataPtr = &myData;
        XR_ProtectSysCall(myData.buf, myData.maxlen);
    } else {
        myDataPtr = 0;
        XR_ProtectSysCall(0,0);
    }
    if ( gma->gma_ctlptr ) {
        myCtl = *(gma->gma_ctlptr);
        myCtlPtr = &myCtl;
        XR_UnprotectSeg(myCtl.buf, myCtl.maxlen);
    } else {
        myCtlPtr = 0;
    }
    if ( gma->gma_flags ) {
        myFlags = *(gma->gma_flags);
        myFlagsPtr = &myFlags;
    } else {
        myFlagsPtr = 0;
    }
    ans = getmsg(fdc->fdc_index, myCtlPtr, myDataPtr, myFlagsPtr);
    XR_UnprotectSysCall();
    if (ans >= 0) {
        if( gma->gma_ctlptr ) gma->gma_ctlptr -> len = myCtl.len;
        if( gma->gma_dataptr ) gma->gma_dataptr -> len = myData.len;
        if( gma->gma_flags ) *(gma->gma_flags) = myFlags;
    }   
    return( ans );
}
#undef gma

#endif


int
XR_GetMsg(fildes, ctlptr, dataptr, flags)
    XR_Fildes fildes;
    struct strbuf *ctlptr;
    struct strbuf *dataptr;
    int *flags;
{
#   if STREAMS
    struct XR_GetMsgArgsRep args;
    XR_UIOWaitReadyProc waitReadyProc;

    args.gma_ctlptr = ctlptr;
    args.gma_dataptr = dataptr;
    args.gma_flags = flags;
    waitReadyProc = &XR_UIOInPri0WaitReadyProc;
    if( (flags) && (flags[0] == RS_HIPRI) )
	waitReadyProc = &XR_UIOPri0WaitReadyProc;
    return(
	XR_UIODoWithFDEAndFDC(
	    fildes,
	    waitReadyProc,
	    XR_GetMsgWorker,
	    ((unsigned)(&args)),
	    0
	)
    );
#   else
	XR_SetErrno(ENOSTR);
	return(-1);
#   endif
}



#if STREAMS

struct XR_PutMsgArgsRep {
    struct strbuf *pma_ctlptr;
    struct strbuf *pma_dataptr;
    int pma_flags;
};


#define pma	((struct XR_PutMsgArgsRep *)(x1))

static XR_FDE_FDC_WORKER(XR_PutMsgWorker)
{
    int ans;

    ans = putmsg(fdc->fdc_index,
	    pma->pma_ctlptr, pma->pma_dataptr, pma->pma_flags);
    return( ans );
}

#undef pma

#endif



int
XR_PutMsg(fildes, ctlptr, dataptr, flags)
    XR_Fildes fildes;
    struct strbuf *ctlptr;
    struct strbuf *dataptr;
    int flags;
{
#   if STREAMS
    struct XR_PutMsgArgsRep args;

    args.pma_ctlptr = ctlptr;
    args.pma_dataptr = dataptr;
    args.pma_flags = flags;
    return(
	XR_UIODoWithFDEAndFDC(
	    fildes,
	    &XR_UIOOut0WaitReadyProc,
	    XR_PutMsgWorker,
	    ((unsigned)(&args)),
	    0
	)
    );
#   else
	XR_SetErrno(ENOSTR);
	return(-1);
#   endif
}

